/*
 * $Id: connection_thread.c,v 1.7 1998/04/07 02:42:59 labovit Exp $
 */

#include <sys/types.h> 
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include <mrt.h>
#include <select.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "salserver.h"

/* globals */
extern trace_t			*default_trace;
extern salserver_t		SALSERVER;

/* local functions */
static int _sal_delete_connection (sal_connection_t *connection);



void schedule_get_archive_config () {
  schedule_event2 ("get_archive_config", SALSERVER.connection_schedule, 
		   get_archive_config, 0, NULL);
}

void schedule_connection_socket_read (sal_connection_t *connection) {
  if (connection->state == 3) {return;}
  schedule_event2 ("connection_socket_read", SALSERVER.connection_schedule, 
		   sal_read, 1, connection);
}

void schedule_connection_write (sal_connection_t *connection) {
  if (connection->state == 3) return;
  schedule_event2 ("connection_write", SALSERVER.connection_schedule, 
		   sal_write, 1, connection);
}


void schedule_connection_tcp_recv () {
  schedule_event2 ("connection_tcp_recv", SALSERVER.connection_schedule, 
		   (void *) connection_tcp_recv, 0);
}


void schedule_show_connections (uii_connection_t *uii) {

#ifdef HAVE_LIBPTHREAD
  pthread_mutex_lock (&uii->mutex_cond_lock);
  schedule_event2 ("show_connections", SALSERVER.connection_schedule, 
		   (void *) show_connections, 1, uii);

  /* wait for connections  to complete */
  pthread_cond_wait (&uii->cond, &uii->mutex_cond_lock);
  pthread_mutex_unlock (&uii->mutex_cond_lock);
#endif /* HAVE_LIBPTHREAD */

  uii_send_bulk_data (uii);
  
  /* free up the memory */
  Delete (uii->answer);
  return;
}



static void _close_salserver_connection (char *host, int port) {
  sal_connection_t *con;
  LL_Iterate (SALSERVER.ll_sal_connections, con) {
    if ((!strcmp (con->name, host)) && (port == con->sockfd)) {
      sal_delete_connection (con);
      return;
    }
  }
}


void close_salserver_connection (uii_connection_t *uii) {
  char host[100];
  int port;

  if (parse_line (uii->cp, "%s %d", host, &port) >= 1) 
    schedule_event2 ("close connection", SALSERVER.connection_schedule, 
		     (void *) _close_salserver_connection, 2, host, port);

  return;
}




void schedule_show_channels (uii_connection_t *uii) {

#ifdef HAVE_LIBPTHREAD
  pthread_mutex_lock (&uii->mutex_cond_lock);

  schedule_event2 ("show channels", SALSERVER.connection_schedule, 
		   (void *) show_channels, 1, uii);

  /* wait for connections  to complete */
  pthread_cond_wait (&uii->cond, &uii->mutex_cond_lock);
  pthread_mutex_unlock (&uii->mutex_cond_lock);
#endif /*  HAVE_LIBPTHREAD */

  uii_send_bulk_data (uii);
  
  /* free up the memory */
  Delete (uii->answer);
  return;
}

/* write out queued up buffers */
void sal_write (sal_connection_t *connection) {
  sal_buffer_t *buffer;
  int n, written;

  SALSERVER.where = "sal_write";

  if (connection->state == 3) {return;}


  while ((more_to_write (connection->sockfd) == 1) && 
	 (LL_GetCount (connection->ll_buffers) > 0)) {

    buffer = LL_GetHead (connection->ll_buffers);
    written = buffer->cp - buffer->data;

    SALSERVER.where = "sal_write:write";

    if (connection->state == 3) {return;}
    if (connection->sockfd <= 0) {return;}

    trace (NORM, default_trace, ".... About to write to %s:%d (%d bytes) (%d quueued)\n",
	   connection->name, connection->sockfd, buffer->size - written, 
	   LL_GetCount (connection->ll_buffers));

    if (more_to_write (connection->sockfd) != 1) {
      trace (NORM, default_trace, ".... *** ABORTING -- not more to write!!!\n");
      break;
    }
    n = write (connection->sockfd, buffer->cp, buffer->size - written);
    buffer->cp += n;

    /* error! */
    if (n <= 0) {
      SALSERVER.where = "sal_write:error";
      trace (NORM, default_trace, "Error writing! %d/%d %s\n", 
	     n,  buffer->size - written, strerror (errno));
      LL_Clear (connection->ll_buffers);
      /* close the durn socket */
      if (connection->multiserver) 
	sal_close_multiserver(connection);
      else
	sal_delete_connection (connection);
      return;
    }

    SALSERVER.where = "sal_write:afterwrite";

    if ((buffer->cp - buffer->data) >= buffer->size) {
      trace (NORM, default_trace, "Yip! another packet out the door... (%d bytes)\n", n);
      LL_Remove (connection->ll_buffers, buffer);

      if (LL_GetCount (connection->ll_buffers) <= 0)
	return; /* don't renable write socket notification */
    }
  }
  
  SALSERVER.where = "sal_write:select_enable";

  /* only enable if more to write */
  if (LL_GetCount (connection->ll_buffers) > 0) 
    select_enable_fd_mask (connection->sockfd, SELECT_WRITE);
}

int sal_process_query (sal_connection_t *connection, data_pdu_t *query) {
  char tmp[512], *name;
  data_pdu_t *reply;
  int fd;
  int persistent_flag = 0; /* flag if we should keep around query */

  if ((connection == NULL)  || (connection->state == 3)) {
    trace (NORM, default_trace, "Skipping -- channel scheduled for deletion\n");
    return (persistent_flag);
  }

  /* Meta Query */
  if (query->flag == SAL_COMMAND_META_QUERY) {
    sal_send_meta_response (connection, query);
    connection->time_last = time (NULL);
    return (persistent_flag);
  }

  name = (char *) sal_get_property (query->ll_properties, "NAME");
  if (name == NULL) {
     trace (NORM, default_trace, "No channel name in query!\n");
     return (-1);
  }

  trace (NORM, default_trace, "Query from %s:%d for %s\n", 
	 connection->name, connection->sockfd, name);
  
  /* query for a specific date */
  if (query->date != NULL) {
    sprintf (tmp, "%s/%s/%s.%s", SALSERVER.dir, query->name, query->name, query->date);
  }
  else if (query->time_closest > 0) {
    reply = time_closest (query, connection);
  }
  /* last */
  else {
    sprintf (tmp, "%s/%s/%s.last", SALSERVER.dir, query->name, query->name);
    persistent_flag = 1;     /* add us to get future push data */
  }
  

  if (query->time_closest <= 0) {
    trace (TR_TRACE, default_trace, "Attempting to open file %s\n", tmp);
    
    if ((fd = open (tmp, O_RDONLY, NULL)) < 0) {
      trace (NORM, default_trace, "No %s last available\n", name);
      /* send back report to client */
      sal_send_error (query->query_id, connection, query->name, "NO LAST AVAILABLE");
      /*close (fd);*/
      return (0);
    } 

    trace (NORM, default_trace, "Reading file %s\n", tmp);
    reply = (data_pdu_t *) sal_read_disk (fd, tmp);
    close (fd);
  }

  if (reply == NULL) {
    trace (NORM, default_trace, "No data read or error for %s\n", name);
    sal_send_error (query->query_id, connection, query->name, "Data Corrupted");
    /*sal_delete_connection (connection); should we really do this? */
    return (1);
  }

  /* save query */
  if (persistent_flag)
    LL_Add (connection->ll_queries, query);

  /* change command field to reply */
  sal_change_property (reply->ll_properties, "COMMAND", "QUERY RESPONSE");

  /* okay , send the data packet */
  sal_send_data (query->query_id, connection, reply);

  /* if we wrote successfully, update the last time for this channel */
  connection->time_last = time (NULL);

  Delete_Data (reply);

  return (1);
}


int sal_send_data (int query_id, sal_connection_t *connection, data_pdu_t *data) {
  char buf[4096];
  int proplen;
  sal_buffer_t *buffer;
  int count;

  SALSERVER.where = "sal_send_data";

  if (connection->state != 1) {
    return (-1);
  }
  
  /* marshall property data */
  sal_marshall_prop (query_id, data->ll_properties, buf, &proplen);

  buffer = (sal_buffer_t *) New_Sal_Buffer (12 + proplen + data->datalen);

  /* header */
  UTIL_PUT_NETLONG (SALSERVER.key, buffer->cp); /* key */
  memcpy (buffer->cp, thisInterfaceVersion, 4); /* interface version */
  buffer->cp +=4;
  UTIL_PUT_NETLONG (proplen, buffer->cp); /* property length */
  
  /* properties */
  memcpy (buffer->cp, buf, proplen);
  buffer->cp += proplen;

  memcpy (buffer->cp, data->databuf, data->datalen);
  buffer->cp = buffer->data;

  /* check if other data already waiting, if so queue this up
   * if too many, kill connection 
   */
  count = LL_GetCount (connection->ll_buffers);
  
  /* too much queued! error! errro! */
  if (count > 100) {
    trace (NORM, default_trace, 
	   "ERROR --  Too much buffered (%d) to %s -- Aborting send...\n", 
	   count, connection->name);    
    sal_delete_connection (connection);
    return (-1);
  }

  if (count > 0) 
    trace (NORM, default_trace, "  Already buffered (%d) to %s -- Queuing data...\n", 
	   count, connection->name);    
  else
    select_enable_fd_mask (connection->sockfd, SELECT_WRITE);

  LL_Append (connection->ll_buffers, buffer);

  
  
  return (0);
}

void sal_process_data (sal_connection_t *sal_connection) {
  data_pdu_t *data;
  sal_connection_t *con;
  sal_connection_t *multiserver;
  char *publish;
  int n;

  SALSERVER.where = "sal_process_data";

  data = sal_connection->data_pdu;
  
  /* save to disk */
  sal_archive (data);
    
  /* add to our list of channels */
  sal_add_channel (data->name);

  /* add to us */
  n = 1;
  LL_Iterate (sal_connection->ll_publish, publish) {
    if (!strcmp (publish, data->name)) {
      n = 0;
      break;
    }
  }
  if (n == 1) {LL_Add (sal_connection->ll_publish, strdup (data->name));}


  SALSERVER.where = "sal_process_data:connections";

  /* 
   * run through current connections and see if anyone wants it 
   */
  LL_Iterate (SALSERVER.ll_sal_connections, con) {
    char *name;
    data_pdu_t *dq;

    SALSERVER.where = "sal_process_data:loop1";

    LL_Iterate (con->ll_queries, dq) {

      SALSERVER.where = "sal_process_data:loop2";

      name = (char *) dq->name;
      if (name == NULL) continue;
      if (!strcmp (name, data->name)) {
	trace (NORM, default_trace, "Sending %s to %s:%d (%d bytes)!\n", 
	       name, con->name, con->sockfd, data->datalen);
	n = sal_send_data (dq->query_id, con, data);
	con->time_last = time (NULL);

	/* error! */
	if (n < 0) {
	  sal_delete_connection (con);
	}
	continue; /* only send once per connection */
      }
    }
  }

  SALSERVER.where = "sal_process_data:multiservers";

  /* 
   * run through multiserver peers and give them data 
   */
  LL_Iterate (SALSERVER.ll_multiserver, multiserver) {
    if ((multiserver->state == SAL_STATE_ESTABLISHED) && (multiserver->sockfd >= 0)) {
      trace (NORM, default_trace, "MULTISERVER Sending %s to %s:%d!\n", data->name, 
	     multiserver->name, multiserver->sockfd);

      if (data->proplen > SALSERVER.max_data_size) {
	trace (NORM, default_trace, "ERROR -- %d bytes mutliserver > %d\n",
	       data->proplen, SALSERVER.max_data_size);
	continue;
      }

      n = sal_send_data (0, multiserver, data);

      if (n < 0) {
	trace (NORM, default_trace, "ERROR -- multiserver write to %s failed!\n", 
	       multiserver->name);
	sal_close_multiserver (multiserver);
      }
    }
    else {
      trace (NORM, default_trace, 
	     "MULTISERVER down -- not sending %s to %s:%d!\n", data->name, 
	     multiserver->name, multiserver->sockfd);
    }
  }

  return;
}


void Delete_Publish (char *publish) {
  Delete (publish);
}


sal_connection_t *New_Sal_Connection (prefix_t *prefix, int fd) {
  sal_connection_t *tmp;

  tmp = New (sal_connection_t);
  tmp->prefix = prefix;
  tmp->sockfd = fd;
  tmp->time_start = time (NULL);
  tmp->time_last = time (NULL);
  tmp->ll_queries = LL_Create (LL_DestroyFunction, Delete_Data, NULL);
  tmp->ll_publish = LL_Create (LL_DestroyFunction, Delete_Publish, NULL);
  tmp->ll_buffers = LL_Create (LL_DestroyFunction, Delete_Sal_Buffer, NULL);
  tmp->data_pdu = (data_pdu_t *) New_Data_Pdu ();
  tmp->name = prefix_toname (prefix);

  return (tmp);
}


sal_connection_t *add_sal_connection (int fd, struct sockaddr_in *addr) {
  prefix_t *prefix;
  sal_connection_t *sal_connection;

  SALSERVER.where = "add_sal_connection";

  if (SALSERVER.num_connections > SALSERVER.max_connections) {
    trace (NORM |TR_ERROR, default_trace, "Too many connections -- rejecting\n");
    close (fd);
    return (NULL);
  }
  SALSERVER.num_connections++;

  prefix = New_Prefix (AF_INET, (char *) &addr->sin_addr, 32);

  sal_connection = New_Sal_Connection (prefix, fd);

  /* make it non-blocking -- for some reason we were hanging in writes even
   * though select said the socket was ready for writing
   * grumble, grumble...
   */
  fcntl (sal_connection->sockfd, F_SETFL, O_NDELAY);

  select_add_fd (fd, SELECT_READ, (void *) schedule_connection_socket_read, 
		 sal_connection);
  select_add_fd_off (fd, SELECT_WRITE, (void *) schedule_connection_write, 
		     sal_connection);
  return (sal_connection);
}



void sal_delete_connection (sal_connection_t *connection) {
  if (connection->state == 3) return;

  connection->state = 3;
  schedule_event2 ("connection_delete", SALSERVER.connection_schedule, 
		   (void *) _sal_delete_connection, 1, connection);
}

/* make it a static so only we can call it */
static int _sal_delete_connection (sal_connection_t *connection) {

  SALSERVER.where = "sal_delete_connection";

  trace (NORM, default_trace, "Closing connection to %s (fd %d)\n", 
	 connection->name, connection->sockfd);
  select_delete_fd (connection->sockfd);

  LL_Remove (SALSERVER.ll_sal_connections, connection);
  Delete_Prefix (connection->prefix);
  Delete (connection->name);
  Delete_Data (connection->data_pdu);
  LL_Destroy (connection->ll_queries);
  LL_Destroy (connection->ll_publish);
  LL_Destroy (connection->ll_buffers);
  Delete (connection);
  SALSERVER.num_connections--;
  return (1);
}

int connection_tcp_recv () {
  int sockfd, len;
  struct sockaddr_in addr;

  SALSERVER.where = "connection_tcp_recv";

  len = sizeof (struct sockaddr_in);

  if ((sockfd = accept (SALSERVER.tcp_sockfd, (struct sockaddr *) &addr, &len)) < 0) {
    trace (NORM, default_trace, "ERROR: %s\n", strerror (errno));
    trace (NORM, default_trace, "Error accepting TCP connetion...\n");
    select_enable_fd (SALSERVER.tcp_sockfd);
    return (-1);
  }

  trace (NORM, default_trace, "Accepted TCP connetion port %d from %s\n", 
	 SALSERVER.tcp_port, inet_ntoa (addr.sin_addr));

  select_enable_fd (SALSERVER.tcp_sockfd); 
  add_sal_connection (sockfd, &addr);

  return (1);
}


void show_channels (uii_connection_t *uii) {
  char *channel;
  char *cp;

  SALSERVER.where = "show_connections";

  cp = uii->answer = malloc (SAL_LARGE_BUF);
  memset (cp, 0, SAL_LARGE_BUF);

  sprintf (cp, "%d Channels\r\n", (int) LL_GetCount (SALSERVER.ll_channels));
  cp += strlen (cp);
  sprintf (cp, "--------------------------\r\n");
  cp += strlen (cp);

  LL_Iterate (SALSERVER.ll_channels,  channel) {
    /*uii_send_data (uii, "  %s\r\n", channel);*/
    sprintf (cp, "  %s\r\n", channel);
    cp += strlen (cp);

    if ((cp - uii->answer) > SAL_LARGE_BUF - 300) {
      sprintf (cp, "  ** Truncated ***\r\n");
      cp += strlen (cp);
      break;
    } 
  }

  /* okay, let the waiting uii know he can print out this buffer */
  signal_uii_data_ready (uii);
} 



/* show_connections
 */
void show_connections (uii_connection_t *uii) {
  sal_connection_t *connection;
  data_pdu_t *query;
  char *publish;

  SALSERVER.where = "show_connections";


  uii_add_bulk_output (uii, "%d Connections [Max: %d, Idle Max %d]\r\n", 
	   SALSERVER.num_connections,
	   SALSERVER.max_connections,
	   SALSERVER.max_idle);
  uii_add_bulk_output (uii, "------------------------------------------------\r\n");

  LL_Iterate (SALSERVER.ll_sal_connections,  connection) {
    if (connection->state == 3) {continue;}
    uii_add_bulk_output (uii, "  %s:%d (Age %d) [%d queued]\r\n", 
			 connection->name, connection->sockfd,
			 time(NULL) - connection->time_last,
			 LL_GetCount (connection->ll_buffers));

    query = NULL;
    LL_Iterate (connection->ll_queries, query) {
      uii_add_bulk_output (uii, "      Query %s\r\n", query->name);
    }
    query = NULL;
    LL_Iterate (connection->ll_publish, publish) {
      uii_add_bulk_output (uii, "      Publish %s\r\n", publish);
    }
  }
    
  signal_uii_data_ready (uii);
  return;
} 

void connection_thread () {
#ifdef HAVE_LIBPTHREAD
  sigset_t set;
  sigemptyset (&set);
  sigaddset (&set, SIGALRM);
  sigaddset (&set, SIGHUP);
  pthread_sigmask (SIG_BLOCK, &set, NULL);
#endif /* HAVE_LIBPTHREAD */
 
  multiserver_retry ();

#ifdef HAVE_LIBPTHREAD
  while (1) {
    SALSERVER.where = "Waiting for event";
    schedule_wait_for_event (SALSERVER.connection_schedule);
  }
  fprintf (stdout, "Exiting PING while loop... should never get here!\n");
  exit (0);
#else
   return;
#endif /* HAVE_LIBPTHREAD */
}


void start_connection_thread () {
  mrt_thread_create ("Connection Thread", SALSERVER.connection_schedule, 
                     (void *) connection_thread, NULL);

}








int sal_process_unsubscribe (sal_connection_t *connection, data_pdu_t *query) {
  data_pdu_t *tmp_query;
  int found;

  if ((connection == NULL)  || (connection->state == 3)) {
    trace (NORM, default_trace, "Skipping -- channel scheduled for deletion\n");
    return (0);
  }

  connection->time_last = time (NULL);

  found = 0;
  LL_Iterate (connection->ll_queries, tmp_query) {
    if (tmp_query->query_id <= 0)
      continue;

    if (query->query_id == tmp_query->query_id) {
      found = 1;
      break;
    }
  }

  if (!found) {
    trace (NORM, default_trace, "No query_id %d to unsubscribe\n", query->query_id);
    sal_send_error (query->query_id, connection, query->name, "Query does not exist");
    return (0);
  }


  LL_Remove (connection->ll_queries, tmp_query);
  return (1);
}
