/*
 * $Id: multiserver.c,v 1.4 1998/03/18 16:18:57 labovit Exp $
 */

#include <sys/types.h> 
#include <sys/stat.h>
#include <fcntl.h>
#include <mrt.h>
#include <select.h>
#include "salserver.h"

/* global functions */
extern trace_t			*default_trace;
extern salserver_t		SALSERVER;

/* local functions */
static int multiserver_connect (sal_connection_t *multiserver);



/* interate through multiserver */
/* if down, try reconnecting */
static void _multiserver_retry () {
  sal_connection_t *multiserver;
  SALSERVER.where = "multiserver_retry";

  LL_Iterate (SALSERVER.ll_multiserver, multiserver) {
    if (multiserver->state != SAL_STATE_NEED_KEY) {
      continue;
    }
    multiserver_connect (multiserver);
  }
  return;
}


void multiserver_retry () {
  schedule_event2 ("multiserver_rety", SALSERVER.connection_schedule, 
		   (void *) _multiserver_retry, 0);
}



/* show multiservers */
void show_multiservers (uii_connection_t *uii) {
  sal_connection_t *multiserver;

  uii_send_data (uii, "Retry timer in %d seconds\r\n", 
		 time_left (SALSERVER.multiserver_retry_timer));

  LL_Iterate (SALSERVER.ll_multiserver,  multiserver) {
    uii_send_data (uii, "  %s:(%d) [%d queue] <state %d>\r\n", 
		   multiserver->name, multiserver->sockfd,
		   LL_GetCount (multiserver->ll_buffers),
		   multiserver->state);
  } 
} 


int sal_close_multiserver (sal_connection_t *multiserver) {
  SALSERVER.where = "sal_close_multiserver";

  trace (NORM, default_trace, "Closing multiserver %s:%d\n", 
	 multiserver->name, multiserver->sockfd);

  select_delete_fd (multiserver->sockfd);
  multiserver->sockfd = 0;
  multiserver->state = SAL_STATE_NEED_KEY;
  LL_Clear (multiserver->ll_buffers);

  return (1);
}






/* 
 * sal_add_multiserver
 */
int sal_add_multiserver (char *name) {
  prefix_t *prefix;
  sal_connection_t *multiserver;

  SALSERVER.where = "sal_add_multiserver";

  

  if ((prefix = string_toprefix (name, default_trace)) == NULL) 
    return (-1);
  
  multiserver = New_Sal_Connection (prefix, 0);
  multiserver->multiserver = 1; /* set flag */
  multiserver->state = SAL_STATE_NEED_KEY;


  
  LL_Add (SALSERVER.ll_multiserver, multiserver);
  return (1);
}


static int multiserver_connect (sal_connection_t *multiserver) {
  int n, sockfd;

  multiserver->sockfd = -1;

  if ((sockfd = socket (AF_INET, SOCK_STREAM, 0)) < 0) {
    trace (NORM, default_trace, "Create socket failed:  %s\n", 
	   strerror(errno));
    multiserver->sockfd = -1;
    return (-1);
  }

  trace (NORM, default_trace, "Attempting multiserver peer connection to %s\n", 
         prefix_toa (multiserver->prefix));

  multiserver->state = SAL_STATE_PENDING; /* pending */
  n = nonblock_connect(default_trace, multiserver->prefix, 9000, sockfd);

  /* failure */
  if (n < 0) {
    multiserver->state = SAL_STATE_NEED_KEY;
    return (-1);
  }
  /* pending */
  if (n == 0) {
    return (0);
  }

  multiserver->sockfd = sockfd;

  if (write (multiserver->sockfd, (char *) &SALSERVER.key, 4) != 4) {
    sal_close_multiserver (multiserver);
  }
  
  trace (NORM, default_trace, "- Multiserver connection to %s:%d\n",
	 prefix_toa (multiserver->prefix), multiserver->sockfd);

  select_add_fd_off (multiserver->sockfd, SELECT_WRITE, 
		     (void *) schedule_connection_write, 
		     multiserver);

  multiserver->state = SAL_STATE_ESTABLISHED;
  return (1);
}


