/*
 * $Id: salserver.h,v 1.6 1998/05/05 16:18:46 labovit Exp $
 */

#define SAL_SERVER_INTERFACE_VERSION       {'2','2','0','1'}

/*
typedef struct _multiserver_t {
  char *name;	
  int sockfd;
  prefix_t *prefix;
} multiserver_t;

*/


typedef struct _sal_archive_t {
  char*	name;
  int	flag;		/* 0 = date, 1 = TIME */
} sal_archive_t;

typedef struct _sal_property_t {
  char	*key;
  char	*value;
} sal_property_t;


typedef struct _salserver_t {
  char		*config_file;
  short		udp_port;
  short		tcp_port;
  int		udp_sockfd;
  int		tcp_sockfd;
  char		*where;			/* for debugging, what routine are we in? */
  u_long	key;
  int		telnet_access_list;	/* access list for inbound telnet connections */
  LINKED_LIST	*ll_sal_connections;	/* linked list of sal TCP connections */
  LINKED_LIST	*ll_multiserver;	/* linked list of multiserver peers */
  LINKED_LIST	*ll_archives;		/* linked list of channels to archive */
  LINKED_LIST	*ll_channels;		/* linked list of available channels */
  LINKED_LIST	*ll_cache;		/* linked list of data_pdu cached    */
  int		max_connections;	/* maxmimum number of connections */
  int		max_idle;		/* how long connection can be idle (seconds) */
  int		num_connections;	/* number of current connections */
  char		*dir;			/* database directory */
  long		max_data_size;		/* how many bytes before we say no */
  schedule_t    *connection_schedule;	/* schedule for connection thread */
  mtimer_t	*multiserver_retry_timer; 
} salserver_t;


typedef struct _sal_buffer_t {
  int size;
  char *cp;
  char *data;
} sal_buffer_t;

typedef struct _data_pdu_t {
  char		*name;
  char		*date;		/* only filled in if a query asks for a specific date */
  u_long	time_closest;   /* only filled in if query asks for specific time */
  prefix_t	*source;
  u_char	flag;		/* SAL_COMMAND_PUBLISH, SAL_COMMAND_QUERY, etc */
  long		time;
  LINKED_LIST	*ll_properties;
  long		datalen;	/* length of data portion */
  long		proplen;	/* length of properties */
  u_char	*cp;		/* tmp pointer to where we are in buffer */
  int		state;		/* 0 = need header, 1 = need prop, 2 = need data */
  u_char	header[12];	
  u_char	*databuf;	
  u_char	*propbuf;	

  u_long	query_id;	/* only used if query */
} data_pdu_t;


#define SAL_COMMAND_PUBLISH	0
#define SAL_COMMAND_QUERY	1
#define SAL_COMMAND_META_QUERY	2
#define SAL_COMMAND_UNSUBSCRIBE 3

typedef struct _sal_connection_t {
  int		multiserver;	/* 1 if this connection is a multiserver */
  char		*name;		/* ascii hostname or IP address */
  int		state;		/* 0 == need to read key, 1 = established 
				 * 3 == deleted, 4 == pending connection */
  long		time_start;	/* time connection was made */
  long		time_last;	/* last time data sent/recvieved */
  prefix_t	*prefix;
  int		sockfd;
  data_pdu_t	*data_pdu;	/* temporary storage */
  LINKED_LIST	*ll_queries;	/* ll of data_pdu_t */
  LINKED_LIST	*ll_publish;	/* ll of of char */
  LINKED_LIST	*ll_buffers;    /* linked list of sal_buffer_t -- data waiting to
				 * be written out on connection */

} sal_connection_t;

#define SAL_STATE_NEED_KEY	0
#define SAL_STATE_ESTABLISHED	1
#define SAL_STATE_DELETED	3
#define SAL_STATE_PENDING	4


#define SAL_SMALL_BUF	4096
#define SAL_LARGE_BUF	16384*50

void Delete_Property (sal_property_t *p);
extern char thisInterfaceVersion[];
void mrt_process_signal ();
void Delete_Data (data_pdu_t *data);
void schedule_connection_socket_read (sal_connection_t *connection);
void sal_connection_process (sal_connection_t *sal_connection);
data_pdu_t *sal_connection_read (int fd, char *source);
int check_passwd (uii_connection_t *uii);
char *sal_get_property (LINKED_LIST *ll_properties, char *property);
data_pdu_t *sal_connection_read (int fd, char *source);
sal_connection_t *add_sal_connection (int fd, struct sockaddr_in *addr);
void show_connections (uii_connection_t *uii);

void schedule_multiserver_retry ();
void show_channels (uii_connection_t *uii);
void multiserver_retry ();
void show_multiservers (uii_connection_t *uii);
int sal_send_data (int query_id, sal_connection_t *connection, data_pdu_t *data);
void schedule_show_connections ();
void show_debug (uii_connection_t *uii);
void sal_read (sal_connection_t *connection);
int sal_close_multiserver (sal_connection_t *multiserver);
int sal_send_meta_response (sal_connection_t *connection, data_pdu_t * query);
int sal_send_error (int query_id, sal_connection_t *connection, char *channel, char *error);
int sal_change_property (LINKED_LIST *ll_properties, char *property, char *newvalue);
data_pdu_t *sal_read_disk (int fd, char *source);
int sal_marshall_prop (int query_id, LINKED_LIST *ll_prop, char *buf, int *n);
data_pdu_t *New_Data_Pdu ();
sal_buffer_t *New_Sal_Buffer (int size);
void Delete_Sal_Buffer (sal_buffer_t *tmp);
void sal_write (sal_connection_t *connection);
void schedule_show_channels (uii_connection_t *uii);
int sal_process_query (sal_connection_t *connection, data_pdu_t *query);
int sal_add_channel (char *channel);

/* network I/O prototypes */
int listen_udp_port ();
int listen_tcp_port ();
int more_to_write (int fd);

/* connection processing prototypes */
void start_connection_thread ();
void schedule_connection_tcp_recv ();
int connection_tcp_recv ();
int sal_add_multiserver (char *name);
sal_connection_t *New_Sal_Connection (prefix_t *prefix, int fd);
void multiserver_connection_failed (sal_connection_t *multiserver);
void multiserver_connection_succeeded (sal_connection_t *multiserver);
void schedule_connection_write (sal_connection_t *connection);
void sal_delete_connection (sal_connection_t *connection);

/* configuration routine prototypes */
void config_create_default ();
int start_config (uii_connection_t *uii);
void config_database (uii_connection_t *uii, char *line);
void config_archive (uii_connection_t *uii, char *file, char *type);
void config_multiserver (uii_connection_t *uii, char *line);
void schedule_get_archive_config ();
void get_archive_config();
void close_salserver_connection (uii_connection_t *uii);

/* database prototypes */
int sal_add_archive (char *channel, int type);
int sal_archive (data_pdu_t *data);
